// ephem.c
#include "ephem.h"
#define DEBUG 0

/***************************************************************************/
/* Get most recent ephemeris data from RINEX 2 navigation file. */
unsigned int GetEphemeris(int sv, FILE *infile, double sec, ephemstruc *E) {
   static char buff[90];
   int i, satinfile;

   int yy, mm, dd, hr, min;
   double t, ss;

   rewind(infile);
   while (getline(buff,infile)!=NULL)
      if (strstr(buff,"END OF HEADER")!=NULL) break;

   if (feof(infile)) return UNDEFINED;

   while (getline(buff,infile)!=NULL)
      if (strpbrk(buff,"0123456789")!=NULL && strstr(buff,"COMMENT")==NULL) 
      {
	 satinfile = atoi(buff);
	 if (sv!=satinfile) 
         {
	    for (i=0;i<7;i++)
	       if (getline(buff,infile)==NULL) return UNDEFINED;
	 } 
         else 
         {
	    buff[22] = '\0';
	    sscanf(buff, "%*d %d %d %d %d %d %lf", &yy, &mm, &dd, &hr, &min, &ss);
	    if (yy<70) yy=yy+2000;
	    else yy=yy+1900;
	    t = gpssec(yy,mm,dd,hr,min,ss);
	    
	    if(DEBUG)
	      fprintf(stderr,"sec:%f y:%d m:%d d:%d h:%d m:%d s:%f  t:%f\n",sec,yy,mm,dd,hr,min,ss,t);

	    /* Treat ephemeris as valid 2 hrs before or 4 hrs after
	       time of ephemeris. */
	    if (fabs(ComputeDiffSec(fmod(sec+3600, SECONDS_IN_WEEK), t)) > 10800) 
            {
	       for (i=0;i<7;i++)
		  if (getline(buff,infile)==NULL) return UNDEFINED;
            } 
            else 
            {
	      
		/* Ephemeris record 2. */
	       if (getline(buff,infile)==NULL) return UNDEFINED;
	       E->Crs = getcol(buff,1);
	       E->deltaN = getcol(buff,2);
	       E->M = getcol(buff,3);

	       /* Ephemeris record 3. */
	       if (getline(buff,infile)==NULL) return UNDEFINED;
	       E->Cuc = getcol(buff,0);
	       E->ecc = getcol(buff,1);
	       E->Cus = getcol(buff,2);
	       E->sqrtA = getcol(buff,3);
	       E->A = E->sqrtA * E->sqrtA;

	       /* Ephemeris record 4. */
	       if (getline(buff,infile)==NULL) return UNDEFINED;
	       E->toe = getcol(buff,0);
	       E->Cic = getcol(buff,1);
	       E->Omega = getcol(buff,2);
	       E->Cis = getcol(buff,3);

	       /* Ephemeris record 5. */
	       if (getline(buff,infile)==NULL) return UNDEFINED;
	       E->incl = getcol(buff,0);
	       E->Crc = getcol(buff,1);
	       E->arg = getcol(buff,2);
	       E->Omegadot = getcol(buff,3);

	       /* Ephemeris record 6. */
	       if (getline(buff,infile)==NULL) return UNDEFINED;
	       E->idot = getcol(buff,0);

	       /* Ephemeris record 7. */
	       if (getline(buff,infile)==NULL) return UNDEFINED;
	       /* Ephemeris record 8. */
	       if (getline(buff,infile)==NULL) return UNDEFINED;

	       /* Set time of applicability (TIC in day of ephemeris epoch). */
	       E->toa = 10UL*(((unsigned int)(t+0.5)) % (unsigned int)SECONDS_IN_DAY);
	       return E->toa;
	    }
	 }
      }

   /* Satellite not found. */
   return UNDEFINED;
}

/***************************************************************************/
/* Get the floating point number in the specified column of a
   RINEX 2 navigation record (format 3X, 4D19.12). */
double getcol(char *r, int c) {
   static char col[20];
   char *ptr;
   int i;

   ptr = r + 3 + c * 19;
   for (i=0;i<19;i++,ptr++) col[i] = *ptr;
   col[19] = '\0';
   return atof(col);
}

/***************************************************************************/
/* Get the next line from the input file (up to 80 characters). */
char *getline(char *buff, FILE *file) {
   return fgets(buff, 82, file);
}

/***************************************************************************
* Orbit calculations.
***************************************************************************/
/* Pick out the proper ephemeris to use for the given satellite. */
int SelectEphemeris(ephemstruc *E, int sv, unsigned int tic,int *refyy, int *refmm, int *refdd, int *refh, int *refm, double *refs ,FILE *infile) 
{
   static ephemstruc etable[MAXSATELLITES];
   static int firsttime = 1;

   int i, n;
   double sec;

   if (firsttime == 1) 
   {
      for(i=0;i<MAXSATELLITES;i++) 
      {
	 etable[i].toa = UNDEFINED;
	 etable[i].toe = 0.0;
	 etable[i].A = 0.0;
	 etable[i].sqrtA = 0.0;
	 etable[i].ecc = 0.0;
	 etable[i].incl = 0.0;
	 etable[i].Omega = 0.0;
	 etable[i].M = 0.0;
	 etable[i].arg = 0.0;
	 etable[i].Omegadot = 0.0;
	 etable[i].idot = 0.0;
	 etable[i].deltaN = 0.0;
	 etable[i].Cus = 0.0;
	 etable[i].Cuc = 0.0;
	 etable[i].Crs = 0.0;
	 etable[i].Crc = 0.0;
	 etable[i].Cis = 0.0;
	 etable[i].Cic = 0.0;
      }
      *E = etable[0];
      firsttime = 0;
   }

   n = sv - 1;
   if (etable[n].toa==UNDEFINED) 
   {
      //need to change this call
     //sec = tictogpssec(tic,);
      sec = tictogpssec(tic,refyy,refmm,refdd,refh,refm,refs);
      if(DEBUG)
	fprintf(stderr,"sv =%d tic %ld sec %f\n",sv,tic,sec);
      if (GetEphemeris(sv, infile, sec, &etable[n]) == UNDEFINED)
	 return FALSE;
   }

   *E = etable[n];
   return TRUE;
}

int SelectEphemeris0(ephemstruc *E, int sv, unsigned int tic, fslhdrstruc0 *H, FILE *infile) 
{
   static ephemstruc etable[MAXSATELLITES];
   static int firsttime = 1;

   int i, n;
   double sec;

   if (firsttime == 1) 
   {
      for(i=0;i<MAXSATELLITES;i++) 
      {
	 etable[i].toa = UNDEFINED;
	 etable[i].toe = 0.0;
	 etable[i].A = 0.0;
	 etable[i].sqrtA = 0.0;
	 etable[i].ecc = 0.0;
	 etable[i].incl = 0.0;
	 etable[i].Omega = 0.0;
	 etable[i].M = 0.0;
	 etable[i].arg = 0.0;
	 etable[i].Omegadot = 0.0;
	 etable[i].idot = 0.0;
	 etable[i].deltaN = 0.0;
	 etable[i].Cus = 0.0;
	 etable[i].Cuc = 0.0;
	 etable[i].Crs = 0.0;
	 etable[i].Crc = 0.0;
	 etable[i].Cis = 0.0;
	 etable[i].Cic = 0.0;
      }
      *E = etable[0];
      firsttime = 0;
   }

   n = sv - 1;
   if (etable[n].toa==UNDEFINED) 
   {
      sec = tictogpssec0(tic, H);
      if(DEBUG)
	fprintf(stderr,"sv =%d tic %ld sec %f\n",sv,tic,sec);
      if (GetEphemeris(sv, infile, sec, &etable[n]) == UNDEFINED)
	 return FALSE;
   }

   *E = etable[n];
   return TRUE;
}


int SelectEphemeris3(ephemstruc *E, int sv, unsigned int tic, fslhdrstruc3 *H, FILE *infile) 
{
   static ephemstruc etable[MAXSATELLITES];
   static int firsttime = 1;

   int i, n;
   double sec;

   if (firsttime == 1) 
   {
      for(i=0;i<MAXSATELLITES;i++) 
      {
	 etable[i].toa = UNDEFINED;
	 etable[i].toe = 0.0;
	 etable[i].A = 0.0;
	 etable[i].sqrtA = 0.0;
	 etable[i].ecc = 0.0;
	 etable[i].incl = 0.0;
	 etable[i].Omega = 0.0;
	 etable[i].M = 0.0;
	 etable[i].arg = 0.0;
	 etable[i].Omegadot = 0.0;
	 etable[i].idot = 0.0;
	 etable[i].deltaN = 0.0;
	 etable[i].Cus = 0.0;
	 etable[i].Cuc = 0.0;
	 etable[i].Crs = 0.0;
	 etable[i].Crc = 0.0;
	 etable[i].Cis = 0.0;
	 etable[i].Cic = 0.0;
      }
      *E = etable[0];
      firsttime = 0;
   }

   n = sv - 1;
   if (etable[n].toa==UNDEFINED) 
   {
      sec = tictogpssec3(tic, H);
      if(DEBUG)
	fprintf(stderr,"sv =%d tic %ld sec %f\n",sv,tic,sec);
      if (GetEphemeris(sv, infile, sec, &etable[n]) == UNDEFINED)
	 return FALSE;
   }

   *E = etable[n];
   return TRUE;
}



int SelectEphemeris4(ephemstruc *E, int sv, unsigned int tic, fslhdrstruc4 *H, FILE *infile) 
{
   static ephemstruc etable[MAXSATELLITES];
   static int firsttime = 1;

   int i, n;
   double sec;

   if (firsttime == 1) 
   {
      for(i=0;i<MAXSATELLITES;i++) 
      {
	 etable[i].toa = UNDEFINED;
	 etable[i].toe = 0.0;
	 etable[i].A = 0.0;
	 etable[i].sqrtA = 0.0;
	 etable[i].ecc = 0.0;
	 etable[i].incl = 0.0;
	 etable[i].Omega = 0.0;
	 etable[i].M = 0.0;
	 etable[i].arg = 0.0;
	 etable[i].Omegadot = 0.0;
	 etable[i].idot = 0.0;
	 etable[i].deltaN = 0.0;
	 etable[i].Cus = 0.0;
	 etable[i].Cuc = 0.0;
	 etable[i].Crs = 0.0;
	 etable[i].Crc = 0.0;
	 etable[i].Cis = 0.0;
	 etable[i].Cic = 0.0;
      }
      *E = etable[0];
      firsttime = 0;
   }

   n = sv - 1;
   if (etable[n].toa==UNDEFINED) 
   {
      sec = tictogpssec4(tic, H);
      if(DEBUG)
	fprintf(stderr,"sv =%d tic %ld sec %f\n",sv,tic,sec);
      if (GetEphemeris(sv, infile, sec, &etable[n]) == UNDEFINED)
	 return FALSE;
   }

   *E = etable[n];
   return TRUE;
}



/**************************************************************************/
/* Calculate GPS satellite ECEF x, y, z position at a given tic (referenced
   to 0000 GPS time) from an ephemeris data block. */
void OrbitCalc(vecstruc *xyz, unsigned int tic, int *refyy, int *refmm, int *refdd, int *refh, int *refm, double *refs , ephemstruc *E) {
   int i;
   double gsec;                                    /* GPS second in week. */
   double sqrt_term;
   double tk, n, n0, Mk, vk, phik;
   double Ek, oldEk = 0.0, sinEk, cosEk;
   double deltaUk, deltaRk, deltaIk;
   double uk, rk, ik, Omegak;
   double xk, yk, S, C;

   /* Compute time from reference epoch (s). */

   //need to change this call
   gsec = tictogpssec(tic,refyy,refmm,refdd,refh,refm,refs);
   tk = ComputeDiffSec(gsec, E->toe);

   /* Compute mean motion (rad/s) and corrected mean motion. */
   if (E->sqrtA==0.0) {
      xyz->x = 0.0;
      xyz->y = 0.0;
      xyz->z = 0.0;
      return;
   }
   n0 = sqrt(MU)/(E->sqrtA * E->sqrtA * E->sqrtA);
   n = n0 + E->deltaN;

   /* Compute mean anomaly (rad). */
   Mk = E->M + n * tk;

   /* Iteratively solve Kepler's equation to compute eccentric anomaly. */
   Ek = Mk;
   for (i=0; i<20; i++) {
      sinEk = sin(Ek);
      Ek = Mk + E->ecc * sinEk;
      if (fabs(Ek - oldEk)<1.0E-14) break;
      oldEk = Ek;
   }
   cosEk = cos(Ek);

   /* Compute true anomaly and argument of latitude. */
   sqrt_term = sqrt(1.0 - E->ecc*E->ecc);
   vk = atan2((sqrt_term*sinEk), (cosEk - E->ecc));
   phik = vk + E->arg;

   /* Compute second harmonic perturbations. */
   S = sin(phik);
   C = cos(phik);
   deltaUk = E->Cus * S + E->Cuc * C;
   deltaRk = E->Crs * S + E->Crc * C;
   deltaIk = E->Cis * S + E->Cic * C;

   /* Compute corrected quantities. */
   uk = phik + deltaUk;
   rk = E->A * (1- E->ecc * cosEk) + deltaRk;
   ik = E->incl + deltaIk + E->idot*tk;
   Omegak = E->Omega + (E->Omegadot - OMEGADOT) * tk - OMEGADOT*E->toe;

   /* Compute satellite position in orbit plane. */
   xk = rk * cos(uk);
   yk = rk * sin(uk);

   /* Rotate to ECEF coordinates. */
   xyz->x = xk * cos(Omegak) - yk * cos(ik) * sin(Omegak);
   xyz->y = xk * sin(Omegak) + yk * cos(ik) * cos(Omegak);
   xyz->z = yk * sin(ik);

   return;
}

void OrbitCalc0(vecstruc *xyz, unsigned int tic, fslhdrstruc0 *H, ephemstruc *E) {
   int i;
   double gsec;                                    /* GPS second in week. */
   double sqrt_term;
   double tk, n, n0, Mk, vk, phik;
   double Ek, oldEk = 0.0, sinEk, cosEk;
   double deltaUk, deltaRk, deltaIk;
   double uk, rk, ik, Omegak;
   double xk, yk, S, C;

   /* Compute time from reference epoch (s). */
   gsec = tictogpssec0(tic, H);
   tk = ComputeDiffSec(gsec, E->toe);

   /* Compute mean motion (rad/s) and corrected mean motion. */
   if (E->sqrtA==0.0) {
      xyz->x = 0.0;
      xyz->y = 0.0;
      xyz->z = 0.0;
      return;
   }
   n0 = sqrt(MU)/(E->sqrtA * E->sqrtA * E->sqrtA);
   n = n0 + E->deltaN;

   /* Compute mean anomaly (rad). */
   Mk = E->M + n * tk;

   /* Iteratively solve Kepler's equation to compute eccentric anomaly. */
   Ek = Mk;
   for (i=0; i<20; i++) {
      sinEk = sin(Ek);
      Ek = Mk + E->ecc * sinEk;
      if (fabs(Ek - oldEk)<1.0E-14) break;
      oldEk = Ek;
   }
   cosEk = cos(Ek);

   /* Compute true anomaly and argument of latitude. */
   sqrt_term = sqrt(1.0 - E->ecc*E->ecc);
   vk = atan2((sqrt_term*sinEk), (cosEk - E->ecc));
   phik = vk + E->arg;

   /* Compute second harmonic perturbations. */
   S = sin(phik);
   C = cos(phik);
   deltaUk = E->Cus * S + E->Cuc * C;
   deltaRk = E->Crs * S + E->Crc * C;
   deltaIk = E->Cis * S + E->Cic * C;

   /* Compute corrected quantities. */
   uk = phik + deltaUk;
   rk = E->A * (1- E->ecc * cosEk) + deltaRk;
   ik = E->incl + deltaIk + E->idot*tk;
   Omegak = E->Omega + (E->Omegadot - OMEGADOT) * tk - OMEGADOT*E->toe;

   /* Compute satellite position in orbit plane. */
   xk = rk * cos(uk);
   yk = rk * sin(uk);

   /* Rotate to ECEF coordinates. */
   xyz->x = xk * cos(Omegak) - yk * cos(ik) * sin(Omegak);
   xyz->y = xk * sin(Omegak) + yk * cos(ik) * cos(Omegak);
   xyz->z = yk * sin(ik);

   return;
}

void OrbitCalc3(vecstruc *xyz, unsigned int tic, fslhdrstruc3 *H, ephemstruc *E) {
   int i;
   double gsec;                                    /* GPS second in week. */
   double sqrt_term;
   double tk, n, n0, Mk, vk, phik;
   double Ek, oldEk = 0.0, sinEk, cosEk;
   double deltaUk, deltaRk, deltaIk;
   double uk, rk, ik, Omegak;
   double xk, yk, S, C;

   /* Compute time from reference epoch (s). */
   gsec = tictogpssec3(tic, H);
   tk = ComputeDiffSec(gsec, E->toe);

   /* Compute mean motion (rad/s) and corrected mean motion. */
   if (E->sqrtA==0.0) {
      xyz->x = 0.0;
      xyz->y = 0.0;
      xyz->z = 0.0;
      return;
   }
   n0 = sqrt(MU)/(E->sqrtA * E->sqrtA * E->sqrtA);
   n = n0 + E->deltaN;

   /* Compute mean anomaly (rad). */
   Mk = E->M + n * tk;

   /* Iteratively solve Kepler's equation to compute eccentric anomaly. */
   Ek = Mk;
   for (i=0; i<20; i++) {
      sinEk = sin(Ek);
      Ek = Mk + E->ecc * sinEk;
      if (fabs(Ek - oldEk)<1.0E-14) break;
      oldEk = Ek;
   }
   cosEk = cos(Ek);

   /* Compute true anomaly and argument of latitude. */
   sqrt_term = sqrt(1.0 - E->ecc*E->ecc);
   vk = atan2((sqrt_term*sinEk), (cosEk - E->ecc));
   phik = vk + E->arg;

   /* Compute second harmonic perturbations. */
   S = sin(phik);
   C = cos(phik);
   deltaUk = E->Cus * S + E->Cuc * C;
   deltaRk = E->Crs * S + E->Crc * C;
   deltaIk = E->Cis * S + E->Cic * C;

   /* Compute corrected quantities. */
   uk = phik + deltaUk;
   rk = E->A * (1- E->ecc * cosEk) + deltaRk;
   ik = E->incl + deltaIk + E->idot*tk;
   Omegak = E->Omega + (E->Omegadot - OMEGADOT) * tk - OMEGADOT*E->toe;

   /* Compute satellite position in orbit plane. */
   xk = rk * cos(uk);
   yk = rk * sin(uk);

   /* Rotate to ECEF coordinates. */
   xyz->x = xk * cos(Omegak) - yk * cos(ik) * sin(Omegak);
   xyz->y = xk * sin(Omegak) + yk * cos(ik) * cos(Omegak);
   xyz->z = yk * sin(ik);

   return;
}


void OrbitCalc4(vecstruc *xyz, unsigned int tic, fslhdrstruc4 *H, ephemstruc *E) {
   int i;
   double gsec;                                    /* GPS second in week. */
   double sqrt_term;
   double tk, n, n0, Mk, vk, phik;
   double Ek, oldEk = 0.0, sinEk, cosEk;
   double deltaUk, deltaRk, deltaIk;
   double uk, rk, ik, Omegak;
   double xk, yk, S, C;

   /* Compute time from reference epoch (s). */
   gsec = tictogpssec4(tic, H);
   tk = ComputeDiffSec(gsec, E->toe);

   /* Compute mean motion (rad/s) and corrected mean motion. */
   if (E->sqrtA==0.0) {
      xyz->x = 0.0;
      xyz->y = 0.0;
      xyz->z = 0.0;
      return;
   }
   n0 = sqrt(MU)/(E->sqrtA * E->sqrtA * E->sqrtA);
   n = n0 + E->deltaN;

   /* Compute mean anomaly (rad). */
   Mk = E->M + n * tk;

   /* Iteratively solve Kepler's equation to compute eccentric anomaly. */
   Ek = Mk;
   for (i=0; i<20; i++) {
      sinEk = sin(Ek);
      Ek = Mk + E->ecc * sinEk;
      if (fabs(Ek - oldEk)<1.0E-14) break;
      oldEk = Ek;
   }
   cosEk = cos(Ek);

   /* Compute true anomaly and argument of latitude. */
   sqrt_term = sqrt(1.0 - E->ecc*E->ecc);
   vk = atan2((sqrt_term*sinEk), (cosEk - E->ecc));
   phik = vk + E->arg;

   /* Compute second harmonic perturbations. */
   S = sin(phik);
   C = cos(phik);
   deltaUk = E->Cus * S + E->Cuc * C;
   deltaRk = E->Crs * S + E->Crc * C;
   deltaIk = E->Cis * S + E->Cic * C;

   /* Compute corrected quantities. */
   uk = phik + deltaUk;
   rk = E->A * (1- E->ecc * cosEk) + deltaRk;
   ik = E->incl + deltaIk + E->idot*tk;
   Omegak = E->Omega + (E->Omegadot - OMEGADOT) * tk - OMEGADOT*E->toe;

   /* Compute satellite position in orbit plane. */
   xk = rk * cos(uk);
   yk = rk * sin(uk);

   /* Rotate to ECEF coordinates. */
   xyz->x = xk * cos(Omegak) - yk * cos(ik) * sin(Omegak);
   xyz->y = xk * sin(Omegak) + yk * cos(ik) * cos(Omegak);
   xyz->z = yk * sin(ik);

   return;
}



/***************************************************************************
* Utility/math routines.
***************************************************************************/
/* Correct a time tic according to the 20ms epoch count. */
unsigned int CorrectTIC(unsigned int tic, unsigned int current_epoch) {
   int tic100ms, diff100ms, epoch100ms;
   unsigned int corrtic;

   if (tic==UNDEFINED) return UNDEFINED;

   tic100ms = tic - 10*(tic/10);
   epoch100ms = current_epoch/5;

   /* Compute signed difference of the two times. */
   diff100ms = epoch100ms - tic100ms;
   if (diff100ms<=-5) diff100ms += 10;
   else if (diff100ms>5) diff100ms -= 10;

   /* Add difference to tic to get corrtic (corrected tic). */
   if (diff100ms<0)
      corrtic = (tic+TICS_IN_DAY-abs(diff100ms))%TICS_IN_DAY;
   else
      corrtic = (tic+diff100ms)%TICS_IN_DAY;

   if (tic>=corrtic+5)
      corrtic += TICS_IN_DAY;
   if (tic<corrtic-5)
      corrtic -= TICS_IN_DAY;

   //my addition
   //if(corrtic < 0)
   //  corrtic +=TICS_IN_DAY;

   /* Now tic is corrected. */
   return corrtic;
}

void CorrectTime(unsigned int tic, unsigned int comptic, fslsatstruc *fsldata) {
  
  


  tic=comptic;

  return;
}


/***************************************************************************/
/* Compute the time difference between two given GPS seconds of the week,
   assuming that they are no more than half a week apart. */
double ComputeDiffSec(double s1, double s2) {
   double d;

   d = s1 - s2;
   if (d < (-SECONDS_IN_WEEK/2.0)) d += SECONDS_IN_WEEK;
   else if (d > SECONDS_IN_WEEK/2.0) d -= SECONDS_IN_WEEK;

   return d;
}

/***************************************************************************/
/* Convert year, month, day, hour, minute, second to GPS second in week.
   Note: does NOT work for years > 2068 (days_elapsed too big for int). */
static int doy[12] = {0,31,59,90,120,151,181,212,243,273,304,334};
double gpssec(int y, int m, int d, int hh, int mm, double ss) {
    int years_elapsed;                                /* Years since 1980. */
    int days_elapsed;             /* Days elapsed since Jan 5/Jan 6, 1980. */
    int leap_days;                   /* Leap days since Jan 5/Jan 6, 1980. */

    years_elapsed = y - 1980;

    leap_days = years_elapsed/4 + 1;
    if ((years_elapsed%4)==0 && m<=2)
	leap_days--;

    days_elapsed = years_elapsed*365 + doy[m-1] + d + leap_days - 6;

    return (double)((days_elapsed%7)*SECONDS_IN_DAY + hh*SECONDS_IN_HOUR
	    + mm*SECONDS_IN_MINUTE) + ss;
}

/***************************************************************************/
/* Convert a given tic value to GPS second within a week, given the
   reference time in the FSL header. */
double tictogpssec0(unsigned int tic, fslhdrstruc0 *H) 
{
   int yy, mm, dd, h, m;
   double s;

   /* Compute tic to time and date. */
   tictotmdat0(tic, H, &yy, &mm, &dd, &h, &m, &s);

   if(DEBUG)
     fprintf(stderr,"tictogpssec -- y:%d m:%d d:%d h:%d m:%d s:%f\n",yy,mm,dd,h,m,s);
         
   /* Compute and return GPS second in week. */
   return gpssec(yy,mm,dd,h,m,s);
}

double tictogpssec3(unsigned int tic, fslhdrstruc3 *H) 
{
   int yy, mm, dd, h, m;
   double s;

   /* Compute tic to time and date. */
   tictotmdat3(tic, H, &yy, &mm, &dd, &h, &m, &s);

   if(DEBUG)
     fprintf(stderr,"tictogpssec -- y:%d m:%d d:%d h:%d m:%d s:%f\n",yy,mm,dd,h,m,s);
         
   /* Compute and return GPS second in week. */
   return gpssec(yy,mm,dd,h,m,s);
}

double tictogpssec4(unsigned int tic, fslhdrstruc4 *H) 
{
   int yy, mm, dd, h, m;
   double s;

   /* Compute tic to time and date. */
   tictotmdat4(tic, H, &yy, &mm, &dd, &h, &m, &s);

   if(DEBUG)
     fprintf(stderr,"tictogpssec -- y:%d m:%d d:%d h:%d m:%d s:%f\n",yy,mm,dd,h,m,s);
         
   /* Compute and return GPS second in week. */
   return gpssec(yy,mm,dd,h,m,s);
}


double tictogpssec(unsigned int tic,int *refyy,int *refmm, int *refdd, int *refh, int *refm, double *refs ) 
{
   int yy, mm, dd, h, m;
   double s;
  
   /* Compute tic to time and date. */
   tictotmdat(tic, refyy,refmm,refdd,refh,refm,refs , &yy, &mm, &dd,&h,&m, &s);

   if(DEBUG)
     fprintf(stderr,"tictogpssec -- y:%d m:%d d:%d h:%d m:%d s:%f\n",yy,mm,dd,h,m,s);
         
   /* Compute and return GPS second in week. */
   return gpssec(yy,mm,dd,h,m,s);
}

/***************************************************************************/
/* Convert a given tic value to time and date. */
static int days_in_month[12] = {31,28,31,30,31,30,31,31,30,31,30,31};
void tictotmdat(unsigned int tic, int *refyy, int *refmm, int *refdd, int *refh, int *refm, double *refs, int *yy, int *mm, int *dd,int *h, int *m, double *s) {
   int max_day;
   unsigned int ttemp;
   double daysec, refsec;

   ttemp = tic;
   *h = (int)(ttemp/TICS_IN_HOUR);
   ttemp -= TICS_IN_HOUR * (unsigned int)(*h);
   *m = (int)(ttemp/TICS_IN_MINUTE);
   ttemp -= TICS_IN_MINUTE * (unsigned int)(*m);
   *s = (double) ttemp/TICS_IN_SECOND;
   daysec = *s + ((double) ((*h * 60) + *m)) * 60.0;

   //*yy = H->yy;
   //*mm = H->mm;
   //*dd = H->dd;
   //refsec = H->s + ((double) ((H->h * 60) + H->m)) * 60.0;

   *yy=*refyy;
   *mm=*refmm;
   *dd=*refdd;
   refsec = *refs + ((double) ((*refh * 60) + *refm)) * 60.0;


   if(DEBUG)
     fprintf(stderr,"tictotmdat -- refsec:%f daysec:%f\n",refsec,daysec);

   /* Add one day to date if time is less than reference time. */
   if (daysec < refsec) 
   {
      if(DEBUG)
	fprintf(stderr,"refsec > daysec");
      max_day = days_in_month[*mm-1];
      if (*mm==2 && *yy%4 == 0 && !(*yy%100==0 && *yy%400 !=0)) 
	max_day++;
      if (++(*dd) > max_day) 
      {
	 *dd = 1;
	 if (++(*mm) > 12) 
         {
	    *mm = 1;
	    (*yy)++;
	 }
      }
   }

   return;
}



void tictotmdat0(unsigned int tic, fslhdrstruc0 *H, int *yy, int *mm, int *dd,
		int *h, int *m, double *s) {
   int max_day;
   unsigned int ttemp;
   double daysec, refsec;

   ttemp = tic;
   *h = (int)(ttemp/TICS_IN_HOUR);
   ttemp -= TICS_IN_HOUR * (unsigned int)(*h);
   *m = (int)(ttemp/TICS_IN_MINUTE);
   ttemp -= TICS_IN_MINUTE * (unsigned int)(*m);
   *s = (double) ttemp/TICS_IN_SECOND;
   daysec = *s + ((double) ((*h * 60) + *m)) * 60.0;

   *yy = H->yy;
   *mm = H->mm;
   *dd = H->dd;
   refsec = H->s + ((double) ((H->h * 60) + H->m)) * 60.0;

   if(DEBUG)
     fprintf(stderr,"tictotmdat -- refsec:%f daysec:%f\n",refsec,daysec);

   /* Add one day to date if time is less than reference time. */
   if (daysec < refsec) 
   {
      if(DEBUG)
	fprintf(stderr,"refsec > daysec");
      max_day = days_in_month[*mm-1];
      if (*mm==2 && *yy%4 == 0 && !(*yy%100==0 && *yy%400 !=0)) 
	max_day++;
      if (++(*dd) > max_day) 
      {
	 *dd = 1;
	 if (++(*mm) > 12) 
         {
	    *mm = 1;
	    (*yy)++;
	 }
      }
   }

   return;
}

void tictotmdat3(unsigned int tic, fslhdrstruc3 *H, int *yy, int *mm, int *dd,
		int *h, int *m, double *s) {
   int max_day;
   unsigned int ttemp;
   double daysec, refsec;

   ttemp = tic;
   *h = (int)(ttemp/TICS_IN_HOUR);
   ttemp -= TICS_IN_HOUR * (unsigned int)(*h);
   *m = (int)(ttemp/TICS_IN_MINUTE);
   ttemp -= TICS_IN_MINUTE * (unsigned int)(*m);
   *s = (double) ttemp/TICS_IN_SECOND;
   daysec = *s + ((double) ((*h * 60) + *m)) * 60.0;

   *yy = H->yy;
   *mm = H->mm;
   *dd = H->dd;
   refsec = H->s + ((double) ((H->h * 60) + H->m)) * 60.0;

   if(DEBUG)
     fprintf(stderr,"tictotmdat -- refsec:%f daysec:%f\n",refsec,daysec);

   /* Add one day to date if time is less than reference time. */
   if (daysec < refsec) 
   {
      if(DEBUG)
	fprintf(stderr,"refsec > daysec");
      max_day = days_in_month[*mm-1];
      if (*mm==2 && *yy%4 == 0 && !(*yy%100==0 && *yy%400 !=0)) 
	max_day++;
      if (++(*dd) > max_day) 
      {
	 *dd = 1;
	 if (++(*mm) > 12) 
         {
	    *mm = 1;
	    (*yy)++;
	 }
      }
   }

   return;
}


void tictotmdat4(unsigned int tic, fslhdrstruc4 *H, int *yy, int *mm, int *dd,
		int *h, int *m, double *s) {
   int max_day;
   unsigned int ttemp;
   double daysec, refsec;

   ttemp = tic;
   *h = (int)(ttemp/TICS_IN_HOUR);
   ttemp -= TICS_IN_HOUR * (unsigned int)(*h);
   *m = (int)(ttemp/TICS_IN_MINUTE);
   ttemp -= TICS_IN_MINUTE * (unsigned int)(*m);
   *s = (double) ttemp/TICS_IN_SECOND;
   daysec = *s + ((double) ((*h * 60) + *m)) * 60.0;

   *yy = H->yy;
   *mm = H->mm;
   *dd = H->dd;
   refsec = H->s + ((double) ((H->h * 60) + H->m)) * 60.0;

   if(DEBUG)
     fprintf(stderr,"tictotmdat -- refsec:%f daysec:%f\n",refsec,daysec);

   /* Add one day to date if time is less than reference time. */
   if (daysec < refsec) 
   {
      if(DEBUG)
	fprintf(stderr,"refsec > daysec");
      max_day = days_in_month[*mm-1];
      if (*mm==2 && *yy%4 == 0 && !(*yy%100==0 && *yy%400 !=0)) 
	max_day++;
      if (++(*dd) > max_day) 
      {
	 *dd = 1;
	 if (++(*mm) > 12) 
         {
	    *mm = 1;
	    (*yy)++;
	 }
      }
   }

   return;
}
