/* List of standard libraries to include. */

#ifndef __EPHEM_H
#define __EPHEM_H

#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fslstructs.h"  // fsl header and sat structures
#include "defines.h"

#define SAMPLES_PER_SECOND 50
#define SAMPLES_PER_MINUTE 3000.0
#define QLENGTH 3000     //Correlation queue is 60 s long.


#define MU 3.986005E14              //WGS-84 Earth gravity (m^3/s^2).
#define OMEGADOT 7.2921151467E-5    //WGS-84 Earth rotation (rad/s).
#define PI 3.1415926535898          //GPS value of pi.
#define L1 1.57542E9                //GPS L1 frequency (Hz).
#define SPEED_OF_LIGHT 2.99792458E8 //Speed of light (m/s).

/* List of error codes for ReadFSLRecord. */
#define ERROR -1
#define OK 0
#define TIME_UNREAD 1
#define MARKERS_UNREAD 2
#define CRC_UNREAD 3
#define ENTRY_UNREAD 4
#define CRC_INVALID -5
#define CHECKSUM_INVALID 6
#define SYNC_LOST -7

/* Controls for lowpass (filter routine). */
#define RESET -1.0        //Any negative power value resets filter.
#define NOISE 0            //Channel used for noise.

/* Satellite ephemeris data. */
typedef struct {
   unsigned long toa;                  /* Time of applicability (tics). */
   double toe;                                    /* Time of epoch (s). */
   double A;                                     /* Semimajor axis (m). */
   double sqrtA;            /* Square root of semimajor axis (sqrt(m)). */
   double ecc;                              /* Eccentricity (unitless). */
   double incl;                          /* Inclination at epoch (rad). */
   double Omega;             /* Long. of ascending node at epoch (rad). */
   double M;                            /* Mean anomaly at epoch (rad). */
   double arg;                            /* Argument of perigee (rad). */
   double Omegadot;                 /* Rate of change of Omega (rad/s). */
   double idot;               /* Rate of change of inclination (rad/s). */
   double deltaN;                    /* Mean motion difference (rad/s). */
   double Cus;                  /* Sine correction, arg. of lat. (rad). */
   double Cuc;                /* Cosine correction, arg. of lat. (rad). */
   double Crs;                          /* Sine correction, radius (m). */
   double Crc;                        /* Cosine correction, radius (m). */
   double Cis;                   /* Sine correction, inclination (rad). */
   double Cic;                 /* Cosine correction, inclination (rad). */
} ephemstruc;

/* Cartesian vector. */
typedef struct {
   double x;                                       /* x coordinate (m). */
   double y;                                       /* y coordinate (m). */
   double z;                                       /* z coordinate (m). */
} vecstruc;

/* Abbreviations to simplify use of unsigned types. */
typedef unsigned char UCHAR;
typedef unsigned int UINT;
typedef unsigned long ULONG;

/* Function prototypes. */
/***************************************************************************
* I/O routines.
***************************************************************************/
//int ReadFSLRecord0(int *status, FILE *infile, ULONG *obstic, ULONG *satmarkers,
//                  UINT PRNTable[], fslsatstruc0 PRNEntries[],
//                  fslsatstruc0 *NoiseEntry);
//int CRCcheck(ULONG t, ULONG s, UCHAR crc);
//int ChecksumCheck(fslsatstruc *rec);
//void PrintStatus(int s);
unsigned int GetEphemeris(int sv, FILE *infile, double sec, ephemstruc *E);
double getcol(char *r, int c);
char *getline(char *buff, FILE *file);
/***************************************************************************
* Orbit calculations.
***************************************************************************/
int SelectEphemeris0(ephemstruc *E, int sv, unsigned int tic, fslhdrstruc0 *H,FILE *infile);
int SelectEphemeris3(ephemstruc *E, int sv, unsigned int tic, fslhdrstruc3 *H,FILE *infile);
int SelectEphemeris4(ephemstruc *E, int sv, unsigned int tic, fslhdrstruc4 *H,FILE *infile);
int SelectEphemeris(ephemstruc *E, int sv, unsigned int tic,int *refyy, int *refmm, int *refdd, int *refh, int *refm, double *refs ,FILE *infile);
void OrbitCalc0(vecstruc *xyz, unsigned int tic, fslhdrstruc0 *H, ephemstruc *E);
void OrbitCalc3(vecstruc *xyz, unsigned int tic, fslhdrstruc3 *H, ephemstruc *E);
void OrbitCalc4(vecstruc *xyz, unsigned int tic, fslhdrstruc4 *H, ephemstruc *E);
void OrbitCalc(vecstruc *xyz, unsigned int tic,int *refyy, int *refmm, int *refdd, int *refh, int *refm, double *refs , ephemstruc *E);
/***************************************************************************
* Utility/math routines.
***************************************************************************/
unsigned int CorrectTIC(unsigned int tic, unsigned int  current_epoch);
double ComputeDiffSec(double s1, double s2);
double gpssec(int y, int m, int d, int hh, int mm, double ss);
double tictogpssec0(unsigned int tic, fslhdrstruc0 *H);
void tictotmdat0(unsigned int tic, fslhdrstruc0 *H, int *yy, int *mm, int *dd,
                int *h, int *m, double *s);
double tictogpssec3(unsigned int tic, fslhdrstruc3 *H);
double tictogpssec4(unsigned int tic, fslhdrstruc4 *H);
void tictotmdat3(unsigned int tic, fslhdrstruc3 *H, int *yy, int *mm, int *dd,
                int *h, int *m, double *s);
void tictotmdat4(unsigned int tic, fslhdrstruc4 *H, int *yy, int *mm, int *dd,
                int *h, int *m, double *s);
double tictogpssec(unsigned int tic,int *refyy,int *refmm, int *refdd, int *refh, int *refm, double *refs  );
void tictotmdat(unsigned int tic,  int *refyy, int *refmm, int *refdd, int *refh, int *refm, double *refs, int *yy, int *mm, int *dd, int *h, int *m, double *s);

/***************************************************************************
* Signal Processing.
***************************************************************************/
double lowpass(double x, int ch);
double lowpassE(double x, int ch);

#endif
