/***********************************************************************
* LSUM4.C  -  Read *.FSL (Fast Scintillation Log) Ver. 4 and *.N (RINEX
*           navigation) files from ScintMon Ver. 3.x and create
*           summary file containing satellite x, y, z, measured Doppler,
*           WBPk samples, no. lost lock flags, S4 (scintillation
*           index), autocorrelation 3 dB width..
*
* 4.01      removed use of numcnt
*
* Author:   Theadore Beach, Cornell University
* Modify:   Brent Ledvina, Cornell University, 2003/11/13
***********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "ephem.h"
#include "fsl.h"

#define NAME "lsum4"
#define VERSION "4.01 (040910)"
#define S4_MASK 0.025
#define DEBUG 0

/**************************************************************Main Program*/

void PrintStatus(int s);

int main(int argc, char *argv[]) {
   ephemstruc Ephem;
   FSL fsl;
		     
   FILE *Nfile, *Outfile;

   // for version 4
   fslhdrstruc4 Hdr;
   fslsatstruc4 PRNEntries[2*MAXCHANNELS];
   fslsatstruc4 NoiseEntry;

   int version = -1;
   int noisepresent=0;
   int filter=0;
   vecstruc satpos;

   char fslname0[13], nname0[13];
   char fslname3[13], nname3[13];
   char ucase[90], outt[13];
   static char yyt[5],mmt[3],ddt[3],*ext=".sum";

   static char *navext0 = ".N", *navext3 = ".nav";
   static char *fslext0 = ".FSL", *fslext3 = ".fsl";
      
   int PRNTable[2*MAXCHANNELS];
   unsigned int numtics[2*MAXSATELLITES], numflags[2*MAXSATELLITES];
   unsigned int NoiseWBPk, SigWBPk;
   unsigned int acorr_width, xcorr_offset;
   unsigned int inq[2*MAXSATELLITES], outq[2*MAXSATELLITES];
   unsigned int missing_n[2*MAXSATELLITES], missing_t[2*MAXSATELLITES];
   unsigned int oldticii[2*MAXSATELLITES], curticii[2*MAXSATELLITES]; 

   int numsats;
   int i, k, l, m, numcnt, n1;
   unsigned int j, n2;
   int ii, prn;
   int synclostcount=0;

   unsigned int firstic=UNDEFINED, hdrtic, obstic;
   unsigned int comptic,currtic;//currticmod;
   unsigned int oldntic=UNDEFINED;
   unsigned int starttic, stoptic;
   unsigned int recordnumber=0UL;
   unsigned int newflag0=0, newflag1=0,prevflag0,prevflag1,oldmarkers0=0,
       oldmarkers1=0,satmarkers0,satmarkers1;
   unsigned int hh, mm;

   long hh0, hh1, mm0, mm1;

   unsigned int slpfindex[2*MAXSATELLITES];
   unsigned int nlpfindex;

   double tempsig;
   double nlpf[QLENGTH], sig[2*MAXSATELLITES][QLENGTH], shpf;
   double slpf[2*MAXSATELLITES][QLENGTH];
   double shpfold[2*MAXSATELLITES];
   double sum[2*MAXSATELLITES], sumsq[2*MAXSATELLITES];
   double lag[2*MAXSATELLITES], var[2*MAXSATELLITES]; 
   double queue[2*MAXSATELLITES][QLENGTH];

   double lag1, lagj, mean1;

   double s4;
   double x=0.0, y=0.0, z=0.0;

   /* Allocate space for some 1-D arrays. */
   // Ver. FSL 3 removed
   //PRNEntries3 = (fslsatstruc *) calloc(MAXCHANNELS, sizeof(fslsatstruc));
   //PRNTable = (unsigned int *) calloc(MAXCHANNELS, sizeof(unsigned int));
   

   /* Initialize other 1-D arrays. */
   for (i=0;i<2*MAXSATELLITES;i++) {
      numflags[i] = 0;
      numtics[i] = 0;

      inq[i] = 0U;
      outq[i] = 0U;
      
      shpfold[i] = 0.0;
      for (j=0;j<5;j++) slpf[i][j] = 0.0;

      sum[i] = 0.0;
      sumsq[i] = 0.0;

      lag[i] = 0.0;
      var[i] = 0.0;
   }
   numcnt=0;

   nlpfindex = 0;
   for (i=0;i<QLENGTH;i++) 
     nlpf[i] = 0.0;
   for (i=0;i<2*MAXSATELLITES;i++) {
     //tmpbuff = (double *) farcalloc(QLENGTH, sizeof(double));
     // why not make tmpbuff an array of 3000 elements...
     //tmpbuff = (double *)malloc(QLENGTH,sizeof(double)); 
     slpfindex[i] = 0;  
     for (j=0;j<QLENGTH;j++) 
     {  
       //tmpbuff[j] = 0.0F;
        queue[i][j] = 0.0;
        slpf[i][j] = 0.0;
     }
   }

   /* Print instructions if called incorrectly. */
   if (argc<3 || argc>5) {
      printf("\n%s Version %s\n", NAME, VERSION);
      printf("Produce summary file from FSL version 4 file and .nav "
	     "file.\n");
      printf("\nUsage:  %s -[f][n] infile [start [stop]] \n\n", NAME);
      puts("\t-f      - filter the raw signal to remove multipath");
      puts("\t-n      - do not filter the raw signal"); 
      puts("\tinfile  - prefix of .fsl and .nav files");
      //puts("\tID      - station ID (e.g. a)");
      puts("\tstart   - UTC start time (hhmm, 24 hr)");
      puts("\tstop    - UTC stop time (hhmm, 24 hr)");
      puts("\nNote: \tThe output filename will be automatically given.");  

      return -1;
   }

   /* Determine if filtering needs to be performed */
  if(strncmp(argv[1],"-f",2)==0){
    filter=1;
    if(DEBUG)
      fprintf(stderr,"filtering is enabled\n");
  }else if(strncmp(argv[1],"-n",2)==0){ 
    filter=0;
    if(DEBUG)
      fprintf(stderr,"no filtering\n");
  }
   /* Get start time (if present). */
   if (argc>3) {
      hh0 = atoi(argv[3])/100;
      mm0 = atoi(argv[3]) - hh0*100;
      if (mm0>59) {
	 printf("\n***Error! Invalid start minute %ld. Must be 0-59.\n\n", mm0);
	 return 1;
      }
      hh0 = hh0%24;
     starttic = ((hh0*60)+mm0)*TICS_IN_MINUTE;
   } else starttic = UNDEFINED;

   /* Get stop time (if present). */
   if (argc>4) {
      hh1 = atoi(argv[4])/100;
      mm1 = atoi(argv[4]) - hh1*100;
      if (mm1>59) {
	 printf("\n***Error! Invalid stop minute %ld. Must be 0-59.\n\n", mm1);
	 return 1;
      }
      hh1 = hh1%24;
      stoptic = ((hh1*60)+mm1)*TICS_IN_MINUTE;
   } else stoptic = UNDEFINED;
   if (stoptic<starttic) stoptic += TICS_IN_DAY;

   /* Construct filenames (limited to 13 characters total). */
   i=0;
   //while ((ucase[i]=toupper(argv[2][i]))!='\0' && i<79) i++;
   while ((ucase[i]=argv[2][i])!='\0' && i<79) i++;
   ucase[79] = '\0';
   sprintf(fslname0, "%s%s", ucase,fslext0);
   sprintf(nname0, "%s%s", ucase,navext0);
   
   sprintf(fslname3, "%s%s", ucase,fslext3);
   sprintf(nname3, "%s%s", ucase,navext3);
   
   /* Open input files. */
   //if ((FSLfile = fopen(fslname, "rb"))==NULL) {
   //   printf("\n***Error! %s could not open .fsl file %s\n\n", NAME,
   //fslname);
   //   exit(1);
   //}

  if(fsl.OpenFSLFile(fslname0) < 0)  {
    if(fsl.OpenFSLFile(fslname3) < 0)  {
      fprintf(stderr,"Error opening    file %s or %s... exiting\n",fslname0,fslname3);
      exit(1);
    }
  }
  if((Nfile = fopen(nname0, "rb"))==NULL) {
    if((Nfile = fopen(nname3, "rb"))==NULL)  {
      printf("\n***Error! %s could not open file %s or %s\n", NAME,nname0,nname3);
      exit(1);
    }
  }
   //Determine FSL version
  if(fsl.ReadVersion(version) < 0) {
    fprintf(stderr,"Error reading file version... exiting\n");
    exit(1);
  }

   //if (!fread(&Hdr3,sizeof(fslhdrstruc3),1,FSLfile)) {
   //   puts("\n***Error! Could not read FSL data header.");
   //   exit(1);
   //}
 
  //verify the version
   if (!(version == FSL_VERSION4)) {
      puts("\n***Error! Incompatible FSL version.");
      printf("Program FSL version: %d.  File version is %d.\n",
	     FSL_VERSION4, version);
      exit(1);
   }
   
   // Read FSL header. 
    if(fsl.ReadHeader(&Hdr) < 0) {
      fprintf(stderr,"Error reading header... exiting\n");
      exit(1);
    }

   /* added to make an output file name */
   // ----------------- poor fix --------
   if(Hdr.yy < 70)
     Hdr.yy += 2000;

    sprintf(yyt,"%4d" ,Hdr.yy);
    sprintf(mmt,"%02d",Hdr.mm);
    sprintf(ddt,"%02d",Hdr.dd); 
  
    sprintf(outt,"%c%c%s%s%c%c%s",yyt[2],yyt[3],mmt,ddt,argv[2][6],argv[2][7],ext);
 
	   
  if ((Outfile=fopen(outt,"w"))==NULL)
       {printf("Could not open outfile %s\n", outt);}

   /* Print summary file header. */
   
     fprintf(Outfile,"%s\n", Hdr.marker_name);
     fprintf(Outfile,"%s\n", Hdr.comment);
     fprintf(Outfile,"%-8.1f %-8.1f %-8.1f\n", Hdr.x/1.E3, Hdr.y/1.E3,
		     Hdr.z/1.E3);
     fprintf(Outfile,"%4d %02d %02d %02d %02d\n", Hdr.yy, Hdr.mm, Hdr.dd,
	  Hdr.h, (int)(Hdr.m + Hdr.s/60.0 + 1));
     fprintf(Outfile,"1\n");        //Number of receivers.
   

   /* Get starting TIC according to header and correct start and stop
      times for data extraction (if necessary). */
  
  hdrtic = (((unsigned int) Hdr.h * 60) + (unsigned int) Hdr.m) * 600 +
	    (unsigned int)(Hdr.s * 10 + 0.5);
  
 
   if (starttic!=UNDEFINED && starttic<hdrtic) {
       if (starttic>(TICS_IN_DAY/2)) starttic = hdrtic;
       else starttic+=TICS_IN_DAY;
   }
   if (stoptic!=UNDEFINED && stoptic<starttic) stoptic += TICS_IN_DAY;

   /* This is for the count of missing signal. */
   for (ii=0;ii<2*MAXSATELLITES;ii++) { 
       oldticii[ii]=starttic;
       curticii[ii]=starttic;
   }
   if(DEBUG)
     printf("hdrtic=%d,obstic=%d,firsttic=%d,starttic=%d,stoptic=%d\n",hdrtic,obstic,firstic,starttic,stoptic);




/* Read in entries for each TIC. */
   while (1) {
     if(synclostcount>10000)
       exit(1);

     
     if((numsats=fsl.ReadFSLData(&PRNTable[0],&obstic,&satmarkers0,&satmarkers1,&PRNEntries[0],&NoiseEntry,&noisepresent,0))< 0) {
	  printf("\n***Error reading input file at record %u!\n",
		recordnumber);
	   PrintStatus(numsats);
	   //printf("numsats=%d sl=%d\n",numsats,synclostcount);
	   if(numsats == SYNC_LOST) {
	     synclostcount++;
	     if(fsl.ReSync())
	       continue;     
	   }
	   else if(numsats == CRC_INVALID) 
	     continue;
	   else if(numsats == 0)
	     exit(1);
           else
	     break;
        }
            
      if (numsats == 0) break;
      recordnumber++;

      /* Save first TIC number for summary. */
      if (firstic==UNDEFINED) {
        firstic = obstic&SYNC_MASK;
	//  else if(recordnumber > 1000)
	//   firstic = obstic&SYNC_MASK;
      }      
      /* Skip remainder of processing and continue on to read next
	 FSL record if there are no entries in record. */
      if (numsats==0 && noisepresent == FALSE) 
	  continue;
     
      /**************************If there *are* entries in record*/
      /* Fix starting TIC number mismatch with header. */
      //if (hdrtic!=firstic)
      //  obstic = (obstic - firstic + hdrtic)%TICS_IN_DAY;
	
     
      /* Times in day less than header start time are taken to
	 be in the next day. */
     
      comptic=obstic%TICS_IN_DAY;
      if (comptic<hdrtic) comptic += TICS_IN_DAY;
      
      currtic=comptic;

      //printf("%d,hdrtic=%ld,obstic=%ld,curtic=%ld,firsttic=%ld,starttic=%d,stoptic=%d\n",recordnumber,hdrtic,obstic,currtic,firstic,starttic,stoptic);

    if(DEBUG)
      printf("%d,hdrtic=%d,obstic=%d,curtic=%d,firsttic=%d,starttic=%d,stoptic=%d\n",recordnumber,hdrtic,obstic,currtic,firstic,starttic,stoptic);


    /* Update noise channel smoothing filter. */
      if(filter) {
        if (noisepresent == FALSE) {
	  //for (j=0;j<5;j++) 
	  nlpf[0] = lowpass(RESET, NOISE);
	  nlpfindex = 0;
	  oldntic = UNDEFINED;
        } 
        else {
	  if (oldntic==UNDEFINED) {
	    for (j=0;j<5;j++)
	      nlpf[j] = lowpass((double) NoiseEntry.WBPk[j], NOISE);
	    nlpfindex = j;
	    oldntic = currtic;
	  } 
          else if (currtic == oldntic+1) {
	    for (j=0;j<5;j++)
	      nlpf[nlpfindex+j] = lowpass((double) NoiseEntry.WBPk[j], NOISE);
	    oldntic = currtic;
	   nlpfindex += j;
	  } 
          else {
	    nlpf[0] = lowpass(RESET, NOISE);
	    oldntic = UNDEFINED;
	    nlpfindex = 0;
	  }
        }
      } 
    

    /* Update signal channel smoothing filters. */
    if(filter){
      prevflag0 = newflag0;
      newflag0 = oldmarkers0 ^ satmarkers0;
      if(newflag0!=0) {// && prevflag == newflag) {
        for (i=1;i<=MAXSATELLITES;i++) {
	  if ((newflag0 & 1)==1) {
	    //for (j=0;j<5;j++)
	    shpfold[i-1] = 0.0;
	    slpf[i-1][0] = lowpass(RESET,i);
	    slpfindex[i-1] = 0;
	    //fprintf(stderr,"resetting prn%d t=%d nf=%X om=%X nm=%X\n",i,currtic,oldmarkers ^ satmarkers, oldmarkers,satmarkers);
          }
	  newflag0 >>= 1;
	}
      }
      prevflag1 = newflag1;
      newflag1 = oldmarkers1 ^ satmarkers1;
      if(newflag1!=0) {// && prevflag == newflag) {
        for (i=MAXSATELLITES+1;i<=2*MAXSATELLITES;i++) {
	  if ((newflag1 & 1)==1) {
	    //for (j=0;j<5;j++)
	    shpfold[i-1] = 0.0;
	    slpf[i-1][0] = lowpass(RESET,i);
	    slpfindex[i-1] = 0;
	    //fprintf(stderr,"resetting prn%d t=%d nf=%X om=%X nm=%X\n",i,currtic,oldmarkers ^ satmarkers, oldmarkers,satmarkers);
          }
	  newflag1 >>= 1;
	}
      }
    }
      
     for (i=0;i<numsats;i++) {
        prn = PRNTable[i];
	ii = prn - 1;
	for (j=0;j<5;j++) { 
	  slpf[ii][slpfindex[ii] + j] = lowpass((double)PRNEntries[i].WBPk[j],prn);
	  sig[ii][slpfindex[ii] + j] = (double)PRNEntries[i].WBPk[j];
	}
	slpfindex[ii] += j;
      }
    oldmarkers0 = satmarkers0;
    oldmarkers1 = satmarkers1;

      /* Take care of time comparisons if start, and possibly stop,
	 time(s) given. */
      if (starttic!=UNDEFINED && currtic<starttic) continue;
      if (stoptic!=UNDEFINED && currtic>stoptic) break;

      /* Update data queues for autocorrelation calculation. */
      
	for (i=0;i<numsats;i++) {
	  ii = PRNTable[i] - 1;
	 
	  for (j=0;j<5;j++) {
	    tempsig = (double) PRNEntries[i].WBPk[j];
	    shpf = tempsig - slpf[ii][j];
	    queue[ii][inq[ii]%QLENGTH] = shpf;
	    inq[ii]=inq[ii]%QLENGTH+1;
	  }
	} 
     


      /* Update flag sums and TIC sums. */
      
       for (i=0;i<numsats;i++) {
	 ii = PRNTable[i] - 1;
	 numtics[ii]=numtics[ii]%600+1;
	 if (PRNEntries[i].flags!=0) numflags[ii]++;

         /* Count number of missing signals in a minute. */

	 // Ver. FSL 2 - expression syntax incorrect ??? best guess employed
	 // if ((curticii[ii]-oldticii[ii]>)5)
	 if ((curticii[ii]-oldticii[ii])>5) {
	   missing_n[ii]++;
           missing_t[ii]=missing_t[ii]+(curticii[ii]-oldticii[ii]);
         }
         oldticii[ii]=curticii[ii];
       }
     
      numcnt++;

    
    // Print records every minute. 
    // this is it.  this is where it all happens
      //printf("currtic=%d %d %d\n",currtic,currtic%TICS_IN_MINUTE,numcnt);

      //if (currtic%TICS_IN_MINUTE==599 || numcnt >= 600) {
      
      if (numcnt >= 600) {    
      /* Update sums for S4 calculation. */
      for (i=0;i<2*MAXSATELLITES;i++) {
	  //if(i==23)
	//printf("XXXXXXXXXXXXXXXXXXXXX %d\n",currtic);
	  
	  if(slpfindex[i] > QLENGTH/2) // > 2000 samples
	  {
	    //fprintf(stderr,"SV%d  slpfindex[i]=%d\n",i+1,slpfindex[i]);
	    if(filter) {
              for (j=0;j<slpfindex[i];j++) {
                sum[i] += slpf[i][j]/((double)slpfindex[i]);
	        if (noisepresent == TRUE)
		  sum[i] -= nlpf[j]/((double)slpfindex[i]);
               
	        shpf = sig[i][j] - slpf[i][j];
	        sumsq[i] += shpf*shpfold[i]/((double)slpfindex[i]);
	        shpfold[i] = shpf;
              }
            } else {
	      for(j=0;j<slpfindex[i];j++){
	        //printf("%d sum[%d]=%f\n",j,i,sum[i]);
		sum[i] += sig[i][j];
                sumsq[i] += sig[i][j]*sig[i][j];
              }
	      sum[i]=sum[i]/((double)slpfindex[i]);
	      sumsq[i]=sumsq[i]/((double)slpfindex[i]);
	    }
          }
        }
      //}


       /* Make and print record header. */
      //currticmod=currtic%TICS_IN_DAY;
         hh = currtic/TICS_IN_HOUR;
	 mm = (currtic-hh*TICS_IN_HOUR)/TICS_IN_MINUTE;
	 
	 //fprintf(stderr,"%02d%02d\n",hh,mm);
	 if (noisepresent == FALSE)
	   NoiseWBPk = 0;
	 else
	   NoiseWBPk = (NoiseEntry.WBPk[0]+NoiseEntry.WBPk[1]+NoiseEntry.WBPk[2]+
	     NoiseEntry.WBPk[3]+NoiseEntry.WBPk[4])/5;
	 fprintf(Outfile,"%02d%02d %2d %5d\n", hh, mm, numsats,NoiseWBPk);

	 /* Calculate record data and add it. */
	 for (i=0;i<numsats;i++) 
         {
	    prn = PRNTable[i];
	    ii = prn - 1;

	    if(DEBUG)
	      fprintf(stderr,"currtic=%d numcnt=%d\n", currtic,numcnt);

	    /* Calculate satellite position (km). */
	    
	      if (SelectEphemeris4(&Ephem, prn, currtic, &Hdr, Nfile)==TRUE) 
	      {
	        if(DEBUG)
		  fprintf(stderr,"selecting ephemeris...\n");
	        OrbitCalc4(&satpos, currtic, &Hdr, &Ephem);
	        x = satpos.x/1000.0;
	        y = satpos.y/1000.0;
	        z = satpos.z/1000.0;
	      } else
	        x = 0.0, y =0.0, z = 0.0;
	    
	    /* Get Doppler shift (Hz) and sample of WBPk. */
		 // Ver. FSL 2
		 //dopp = -PRNEntries3[i].range_rate * L1/SPEED_OF_LIGHT;
	    SigWBPk = (PRNEntries[i].WBPk[0]+PRNEntries[i].WBPk[1]+PRNEntries[i].WBPk[2]+
	      PRNEntries[i].WBPk[3]+PRNEntries[i].WBPk[4])/5;
	   

	    /* Calculate S4 (unitless). */
	    if (sumsq[ii]<0 || inq[ii]<2000) {
	       s4 = 0.0;
	       //fprintf(stderr,"PRN %d,inq=%d\n",ii+1,inq[ii]);
	    }
	    else {
	      if(filter)
	         s4 = sqrt(sumsq[ii]/(sum[ii]*sum[ii]))*sqrt((double)slpfindex[ii]/3000.);
              else {
                s4=sqrt(sumsq[ii]-(sum[ii]*sum[ii]))/sum[ii]*sqrt((double)slpfindex[ii]/3000.);
           	//s4=sqrt(sumsq[ii]-(sum[ii]*sum[ii]))/sum[ii];
	      }
            }
	    //if(ii==23)
	    //printf("slpfi=%d numcnt=%d currtic=%d\n",slpfindex[ii],numcnt,currtic);

	    /* Calculate autocorrelation 3 dB width (samples). */
	    acorr_width = 9999;

            if (s4>S4_MASK && inq[ii]>2000) {
             for(n1=0; n1<5; n1++) 
               for(n2=1; n2<inq[ii]-1; n2++)
		  queue[ii][n2]=(queue[ii][n2-1]+queue[ii][n2]+
				 queue[ii][n2+1])/3;
		       
	       mean1=0.0;                  
             for(n2=0; n2<inq[ii]; n2++)
               mean1+=queue[ii][n2]/inq[ii]; 
	    
             for(n2=0; n2<inq[ii]; n2++)
	       queue[ii][n2]=queue[ii][n2]-mean1; 

	       lag1 = 0.0;
               for (k=0,l=inq[ii]-1; k<1000; k++,l=l-1) {
		  lag1 += queue[ii][l]*queue[ii][l];
	       }

               for (j=2;j<=1000;j++) {
		  lagj = 0.0F;
                  for (k=0,l=inq[ii]-1; k<1000; k++,l=l-1) {
		     m = l-j;
		     lagj += queue[ii][l]*queue[ii][m];
		  }
		  
		  if (lagj <= lag1/2) {
		     acorr_width = j;
		     break;
		  }
	       }
	    } 

	    /* Calculate cross correlation offset (samples). */
	    xcorr_offset = 9999;  

	    if(DEBUG)
	      fprintf(stderr,"%2d %-8.1f %-8.1f %-8.1f%5d\n",prn,x,y,z,SigWBPk);

	    /* Output record line. */
	    fprintf(Outfile,"%2d %-8.1f %-8.1f %-8.1f %5u %5u %2u %5.3f %4d %4d\n", prn, x, y, z, missing_n[ii], SigWBPk, missing_t[ii], s4, acorr_width, xcorr_offset);
	    	    
	 }
      }  /* End of writing loop every minute (IF) */

   //printf("\\");
      
      //if (currtic%TICS_IN_MINUTE==599 || numcnt>=600) {  

      if(numcnt>=600) {
	 for (i=0;i<2*MAXSATELLITES;i++) {
            numflags[i]  = 0;
            numtics[i]   = 0;
            inq[i]       = 0;
            sum[i]       = 0.0;
            sumsq[i]     = 0.0;
            missing_n[i] = 0;
            missing_t[i] = 0;
	    slpfindex[i] = 0;
	    for(j=0;j<QLENGTH;j++) {
	      sig[i][j]=0.0;
            }
	 }
	 nlpfindex = 0;
	 numcnt=0;
      }
   }   /* End of reading loop. */

   /* Exit program gracefully. */
   fsl.CloseFile();
   fclose(Nfile);
 
   return 0;
}

/***************************************************************************/
/* Print status after FSL file reading error. */
void PrintStatus(int s) {
   if (s==TIME_UNREAD) puts("Failed to read time word.");
   else if (s==MARKERS_UNREAD) puts("Failed to read satellite markers.");
   else if (s==CRC_UNREAD) puts("Failed to read record header CRC.");
   else if (s==ENTRY_UNREAD) puts("Failed to read a satellite entry.");
   else if (s==CRC_INVALID) puts("Invalid CRC for record header.");
   else if (s==CHECKSUM_INVALID) puts("Invalid checksum for satellite entry.");
   else if (s==SYNC_LOST) puts("Sync word not found in record header.");
   return;
}

/***************************************************************************
* Signal Processing.
***************************************************************************/
/* Low pass filter function with cutoff frequency fco=0.004(f/2),
   where f is the sampling frequency.  For f=50 Hz, fco=0.1 Hz.  The
   filter is a third order Chebyshev filter designed in MATLAB with
   the following command:  [b,a]=cheby1(3, 0.5, 0.004).  Filter
   computations are in "transposed direct form II". */
double lowpass(double x, int ch) {
   #define FILTORDER 3
   #define NUMCHANNELS (2*MAXSATELLITES+1)  //Satellite and noise channels.
   int i, j;
   static int firsttime=TRUE;
   double xn=(double) x, yn;

   /* Filter coefficients: a for y, b for x. */
   static double a[FILTORDER+1] = {1.0, -2.98413761039136, 2.96851781402612,
			       -0.98437879453335};
   static double b[FILTORDER+1] = {1.7613767741564E-7, 5.2841303224693E-7,
			       5.2841303224693E-7, 1.7613767741564E-7};

   /* Buffer for data. */
   static double dbuff[FILTORDER][NUMCHANNELS];

   /* Initialize all buffers to zero the first time through. */
   if (firsttime==TRUE) {
      for (j=0; j<NUMCHANNELS; j++)
	 for (i=0; i<FILTORDER; i++) dbuff[i][j] = (double) 0.0;
      firsttime=FALSE;
   }

   /* Initialize buffer for channel ch if xn<0 (power is always>=0). */
   if (xn<0) {
      for (i=0; i<FILTORDER; i++) dbuff[i][ch] = (double) 0.0;
      return (double) 0.0;
   }

   /* Filter the selected satellite data. */
   yn = b[0]*xn + dbuff[0][ch];
   for (i=0; i<FILTORDER; i++) {
      if (i<FILTORDER-1)
	 dbuff[i][ch] = b[i+1]*xn - a[i+1]*yn + dbuff[i+1][ch];
      else
	 dbuff[i][ch] = b[i+1]*xn - a[i+1]*yn;
   }

   return (double) yn;
}

double lowpassE(double x, int ch) {
   #define FILTORDER 3
   #define NUMCHANNELS (2*MAXSATELLITES+1)  //Satellite and noise channels.
   int i, j;
   static int firsttime=TRUE;
   double xn=(double) x, yn;

   /* Filter coefficients: a for y, b for x. */
   static double a[FILTORDER+1] = {1.0, -2.98413761039136, 2.96851781402612,
			       -0.98437879453335};
   static double b[FILTORDER+1] = {1.7613767741564E-7, 5.2841303224693E-7,
			       5.2841303224693E-7, 1.7613767741564E-7};

   /* Buffer for data. */
   static double dbuff[FILTORDER][NUMCHANNELS];

   /* Initialize all buffers to zero the first time through. */
   if (firsttime==TRUE) {
      for (j=0; j<NUMCHANNELS; j++)
	 for (i=0; i<FILTORDER; i++) dbuff[i][j] = (double) 0.0;
      firsttime=FALSE;
   }

   /* Initialize buffer for channel ch if xn<0 (power is always>=0). */
   if (xn<0) {
      for (i=0; i<FILTORDER; i++) dbuff[i][ch] = (double) 0.0;
      return (double) 0.0;
   }

   /* Filter the selected satellite data. */
   yn = b[0]*xn + dbuff[0][ch];
   for (i=0; i<FILTORDER; i++) {
      if (i<FILTORDER-1)
	 dbuff[i][ch] = b[i+1]*xn - a[i+1]*yn + dbuff[i+1][ch];
      else
	 dbuff[i][ch] = b[i+1]*xn - a[i+1]*yn;
   }

   return (double) yn;
}

