/* List of standard libraries to include. */
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "fsl.h"

/* List of PC specific libraries to include. */
#include <alloc.h>

/* List of definitions. */
#define TRUE 1
#define FALSE 0

#define MAXSATELLITES 32         //Maximum number of satellites allowed.
#define MAXCHANNELS 12           //Maximum number of channels per record.


#define UNDEFINED 0xFFFFFFFFUL   //Value for undefined TIC numbers.
#define TICS_IN_DAY 864000UL
#define TICS_IN_HOUR 36000UL
#define TICS_IN_MINUTE 600UL
#define TICS_IN_SECOND 10.0

#define SECONDS_IN_WEEK 604800.0
#define SECONDS_IN_DAY 86400.0
#define SECONDS_IN_HOUR 3600.0
#define SECONDS_IN_MINUTE 60.0

#define SAMPLES_PER_SECOND 50
#define SAMPLES_PER_MINUTE 3000.0
#define QLENGTH 3000                //Correlation queue is 60 s long.

#define MU 3.986005E14              //WGS-84 Earth gravity (m^3/s^2).
#define OMEGADOT 7.2921151467E-5    //WGS-84 Earth rotation (rad/s).
#define PI 3.1415926535898          //GPS value of pi.
#define L1 1.57542E9                //GPS L1 frequency (Hz).
#define SPEED_OF_LIGHT 2.99792458E8 //Speed of light (m/s).

/* List of error codes for ReadFSLRecord. */
#define ERROR -1
#define OK 0
#define TIME_UNREAD 1
#define MARKERS_UNREAD 2
#define CRC_UNREAD 3
#define ENTRY_UNREAD 4
#define CRC_INVALID 5
#define CHECKSUM_INVALID 6
#define SYNC_LOST 7

/* Controls for lowpass (filter routine). */
#define RESET -1.0F        //Any negative power value resets filter.
#define NOISE 0            //Channel used for noise.


/* Satellite ephemeris data. */
typedef struct {
   unsigned long toa;                  /* Time of applicability (tics). */
   double toe;                                    /* Time of epoch (s). */
   double A;                                     /* Semimajor axis (m). */
   double sqrtA;            /* Square root of semimajor axis (sqrt(m)). */
   double ecc;                              /* Eccentricity (unitless). */
   double incl;                          /* Inclination at epoch (rad). */
   double Omega;             /* Long. of ascending node at epoch (rad). */
   double M;                            /* Mean anomaly at epoch (rad). */
   double arg;                            /* Argument of perigee (rad). */
   double Omegadot;                 /* Rate of change of Omega (rad/s). */
   double idot;               /* Rate of change of inclination (rad/s). */
   double deltaN;                    /* Mean motion difference (rad/s). */
   double Cus;                  /* Sine correction, arg. of lat. (rad). */
   double Cuc;                /* Cosine correction, arg. of lat. (rad). */
   double Crs;                          /* Sine correction, radius (m). */
   double Crc;                        /* Cosine correction, radius (m). */
   double Cis;                   /* Sine correction, inclination (rad). */
   double Cic;                 /* Cosine correction, inclination (rad). */
} ephemstruc;

/* Cartesian vector. */
typedef struct {
   double x;                                       /* x coordinate (m). */
   double y;                                       /* y coordinate (m). */
   double z;                                       /* z coordinate (m). */
} vecstruc;

/* Abbreviations to simplify use of unsigned types. */
typedef unsigned char uchar;
typedef unsigned int uint;
typedef unsigned long ulong;

/* Function prototypes. */
/***************************************************************************
* I/O routines.
***************************************************************************/
int ReadFSLRecord(int *status, FILE *infile, ulong *obstic, ulong *satmarkers,
                  uint PRNTable[], fslsatstruc PRNEntries[],
                  fslsatstruc *NoiseEntry);
void PrintStatus(int s);
ulong GetEphemeris(int sv, FILE *infile, double sec, ephemstruc *E);
double getcol(char *r, int c);
char *getline(char *buff, FILE *file);
/***************************************************************************
* Orbit calculations.
***************************************************************************/
int SelectEphemeris(ephemstruc *E, int sv, ulong tic, fslhdrstruc *H,
                    FILE *infile);
void OrbitCalc(vecstruc *xyz, ulong tic, fslhdrstruc *H, ephemstruc *E);
/***************************************************************************
* Utility/math routines.
***************************************************************************/
ulong CorrectTIC(ulong tic, uint current_epoch);
double ComputeDiffSec(double s1, double s2);
double gpssec(int y, int m, int d, int hh, int mm, double ss);
double tictogpssec(ulong tic, fslhdrstruc *H);
void tictotmdat(ulong tic, fslhdrstruc *H, int *yy, int *mm, int *dd,
                int *h, int *m, double *s);
/***************************************************************************
* Signal Processing.
***************************************************************************/
float lowpass(float x, int ch);
float lowpassE(float x, int ch);
