/* List of standard libraries to include. */

#ifndef __SUM_H
#define __SUM_H

#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* List of PC specific libraries to include. */
//#include <alloc.h>

/* List of definitions. */
#define TRUE 1
#define FALSE 0

#define FSL_VERSION 3            //FSL version number.
#define MAXSATELLITES 32         //Maximum number of satellites allowed.
#define MAXCHANNELS 12           //Maximum number of channels per record.
                                
/* Constants for FSL read routines. */
#define SYNC_MSB 0x57000000UL
#define SYNC_MSB_BYTE 0x57
#define SYNC_MASK 0x00FFFFFFUL
#define FLAG_MASK 0x00FF

#define UNDEFINED 0xFFFFFFFFUL   //Value for undefined TIC numbers.
#define TICS_IN_DAY 864000UL
#define TICS_IN_HOUR 36000UL
#define TICS_IN_MINUTE 600UL
#define TICS_IN_SECOND 10.0

#define SECONDS_IN_WEEK 604800.0
#define SECONDS_IN_DAY 86400.0
#define SECONDS_IN_HOUR 3600.0
#define SECONDS_IN_MINUTE 60.0

#define SAMPLES_PER_SECOND 50
#define SAMPLES_PER_MINUTE 3000.0
#define QLENGTH 3000                //Correlation queue is 60 s long.

#define MU 3.986005E14              //WGS-84 Earth gravity (m^3/s^2).
#define OMEGADOT 7.2921151467E-5    //WGS-84 Earth rotation (rad/s).
#define PI 3.1415926535898          //GPS value of pi.
#define L1 1.57542E9                //GPS L1 frequency (Hz).
#define SPEED_OF_LIGHT 2.99792458E8 //Speed of light (m/s).

/* List of error codes for ReadFSLRecord. */
#define ERROR -1
#define OK 0
#define TIME_UNREAD 1
#define MARKERS_UNREAD 2
#define CRC_UNREAD 3
#define ENTRY_UNREAD 4
#define CRC_INVALID 5
#define CHECKSUM_INVALID 6
#define SYNC_LOST 7

/* Controls for lowpass (filter routine). */
#define RESET -1.0F        //Any negative power value resets filter.
#define NOISE 0            //Channel used for noise.

/* Set up special structures for data read from files. */
/* Header information for scintillation log file. */
typedef struct {
   int FSLversion;                               /* FSL version number. */
   char pgm[21];                             /* Program name for RINEX. */
   char run_by[21];                        /* Run by listing for RINEX. */
   char date[21];                                    /* Date for RINEX. */
   char comment[61];                              /* Comment for RINEX. */
   char marker_name[61];                      /* Marker Name for RINEX. */
   char observer[21];                    /* Observer listing for RINEX. */
   char agency[21];                        /* Agency listing for RINEX. */
   char rec_no[21];                       /* Receiver number for RINEX. */
   char rec_type[21];                       /* Receiver type for RINEX. */
   char rec_vers[21];                    /* Receiver version for RINEX. */
   char ant_no[21];                        /* Antenna number for RINEX. */
   char ant_type[21];                        /* Antenna type for RINEX. */
   double x;                                   /* X-position for RINEX. */
   double y;                                   /* Y-position for RINEX. */
   double z;                                   /* Z-position for RINEX. */
   float delta_h;                          /* Antenna height for RINEX. */
   float delta_e;                            /* Antenna east for RINEX. */
   float delta_n;                           /* Antenna north for RINEX. */
   int yy;                                  /* Starting year for RINEX. */
   int mm;                                 /* Starting month for RINEX. */
   int dd;                                   /* Starting day for RINEX. */
   int h;                                   /* Starting hour for RINEX. */
   int m;                                 /* Starting minute for RINEX. */
   double s;                              /* Starting second for RINEX. */
} fslhdrstruc;

/* Satellite entry in scintillation log file. */
// version 3.0 changed the fsl structure
typedef struct {
  //double pseudorange;                              /* Pseudorange (m). */
  //double range_rate;                        /* Pseudorange rate (m/s). */
   double int_carrier_phase;            /* Carrier beat phase (cycles). */
  //unsigned int NBPk[5];         /* Correlator based narrow band power. */
   unsigned short int WBPk[5];           /* Correlator based wide band power. */
   unsigned short int epoch;           /* Time of first SNR estimator sample. */
   unsigned short int flags;          /* Checksum and flags: e.g., lost lock. */
} fslsatstruc;

/* Satellite ephemeris data. */
typedef struct {
   unsigned long toa;                  /* Time of applicability (tics). */
   double toe;                                    /* Time of epoch (s). */
   double A;                                     /* Semimajor axis (m). */
   double sqrtA;            /* Square root of semimajor axis (sqrt(m)). */
   double ecc;                              /* Eccentricity (unitless). */
   double incl;                          /* Inclination at epoch (rad). */
   double Omega;             /* Long. of ascending node at epoch (rad). */
   double M;                            /* Mean anomaly at epoch (rad). */
   double arg;                            /* Argument of perigee (rad). */
   double Omegadot;                 /* Rate of change of Omega (rad/s). */
   double idot;               /* Rate of change of inclination (rad/s). */
   double deltaN;                    /* Mean motion difference (rad/s). */
   double Cus;                  /* Sine correction, arg. of lat. (rad). */
   double Cuc;                /* Cosine correction, arg. of lat. (rad). */
   double Crs;                          /* Sine correction, radius (m). */
   double Crc;                        /* Cosine correction, radius (m). */
   double Cis;                   /* Sine correction, inclination (rad). */
   double Cic;                 /* Cosine correction, inclination (rad). */
} ephemstruc;

/* Cartesian vector. */
typedef struct {
   double x;                                       /* x coordinate (m). */
   double y;                                       /* y coordinate (m). */
   double z;                                       /* z coordinate (m). */
} vecstruc;

/* Abbreviations to simplify use of unsigned types. */
typedef unsigned char UCHAR;
typedef unsigned int UINT;
typedef unsigned long ULONG;

/* Function prototypes. */
/***************************************************************************
* I/O routines.
***************************************************************************/
int ReadFSLRecord(int *status, FILE *infile, ULONG *obstic, ULONG *satmarkers,
                  UINT PRNTable[], fslsatstruc PRNEntries[],
                  fslsatstruc *NoiseEntry);
int CRCcheck(ULONG t, ULONG s, UCHAR crc);
int ChecksumCheck(fslsatstruc *rec);
void PrintStatus(int s);
ULONG GetEphemeris(int sv, FILE *infile, double sec, ephemstruc *E);
double getcol(char *r, int c);
char *getline(char *buff, FILE *file);
/***************************************************************************
* Orbit calculations.
***************************************************************************/
int SelectEphemeris(ephemstruc *E, int sv, ULONG tic, fslhdrstruc *H,
                    FILE *infile);
void OrbitCalc(vecstruc *xyz, ULONG tic, fslhdrstruc *H, ephemstruc *E);
/***************************************************************************
* Utility/math routines.
***************************************************************************/
ULONG CorrectTIC(ULONG tic, UINT current_epoch);
double ComputeDiffSec(double s1, double s2);
double gpssec(int y, int m, int d, int hh, int mm, double ss);
double tictogpssec(ULONG tic, fslhdrstruc *H);
void tictotmdat(ULONG tic, fslhdrstruc *H, int *yy, int *mm, int *dd,
                int *h, int *m, double *s);
/***************************************************************************
* Signal Processing.
***************************************************************************/
float lowpass(float x, int ch);
float lowpassE(float x, int ch);


#endif
